"""Module for orbital manuevers

Extended description of the module

Notes
-----
    This is an example of an indented section. It's like any other section,
    but the body is indented to help it stand out from surrounding text.

If a section is indented, then a section break is created by
resuming unindented text.

Attributes
----------
module_level_variable1 : int
    Descrption of the variable

Author
------
Shankar Kulumani		GWU		skulumani@gwu.edu
"""

import numpy as np
from kinematics import attitude
import pdb
from scipy.optimize import fsolve
# TODO: Add documentation
from . import kepler

def rvfpa2orbit_el(mag_r, mag_v, fpa, mu):
    """Converts R,V, and FPA to orbital elements

    a, p, ecc, nu = rvfpa2orbit_el(mag_r, mag_v, fpa, mu)

    Inputs: 
        - List all inputs into function

    Outputs: 
        - List/describe outputs of function

    Dependencies: 
        - list dependent m files required for function

    Author: 
        - Shankar Kulumani 7 Oct 2012
        - Shankar Kulumani 5 Nov 2012
            - added ascending/descending logic to true anomaly
        - Shankar Kulumani 16 November 2017
            - move to python
    References
        - AAE532_PS6.pdf
        """
    # calculate semi major axis using energy relationship
    a = -mu / (mag_v**2 / 2 - mu / mag_r) / 2
    sme = -mu / (2 * a)

    # find semi latus rectum and eccentricity
    mag_h = mag_r * mag_v * np.cos(fpa)
    p = mag_h**2 / mu

    ecc = np.sqrt(-p / a + 1)

    # find true anomaly using conic equation
    nu = np.arccos((p / mag_r - 1) / ecc)
    if fpa < 0:
        nu = - nu
    nu = attitude.normalize(nu, 0, 2 * np.pi)[0]
    # orbit_el(p,ecc,0,0,0,nu,mu,'true')

    return a, p, ecc, nu


def single_impulse(mag_r1, mag_v1, fpa1, delta_v, alpha):
    """Given a deltaV and alpha angle outputs the new orbital elements

    mag_rf, mag_vf, fpaf = single_impulse(mag_r1, mag_v1, fpa1, delta_v, alpha)      

   Inputs: 
       - List all inputs into function

   Outputs: 
       - List/describe outputs of function

   Dependencies: 
       - list dependent m files required for function

   Author: 
       - Shankar Kulumani 7 Oct 2012
           - list revisions

   References
       - AAE532_PS6.pdf  
    """
    beta = np.pi - alpha
    mag_vf = np.sqrt(mag_v1**2 + delta_v**2 - 2 *
                     mag_v1 * delta_v * np.cos(beta))

    dfpa_cos = (delta_v**2 - mag_vf**2 - mag_v1**2) / (-2 * mag_vf * mag_v1)
    dfpa_sin = delta_v * np.sin(beta) / mag_vf
    delta_fpa = np.arctan2(dfpa_sin, dfpa_cos)

    mag_rf = mag_r1

    fpaf = fpa1 + delta_fpa

    return mag_rf, mag_vf, fpaf


def nu_solve(r, v, fpa, mu):
    """Single impulse manuever true anomaly 
        nu = nu_solve(r,v,fpa)

     Inputs: 
        - r - magnitude of position vector in km
        - v - magnitude of velocity vector in km
        - fpa - flight path angle in rad -pi/2 < fpa < pi/2
        - mu - gravitational parameter km^3/sec^2

     Outputs: 
        - nu - true anomaly of body in rad 0 < nu < 2*pi

     Dependencies: 
        - none

     Author: 
        - Shankar Kulumani 10 Oct 2012
        - Shankar Kulumani 18 November 2017 
            - in Python

    References
        - AAE532 notes LSN 18
        - AAE532_PS7.pdf
    """
    num = (r * v**2 / mu) * np.cos(fpa) * np.sin(fpa)
    den = (r * v**2 / mu) * np.cos(fpa)**2 - 1

    nu = np.arctan2(num, den)

    nu = attitude.normalize(nu, 0, 2*np.pi)
    return nu

def delta_v_solve_planar(v1, v2, fpa1, fpa2):
    """Solve for DeltaV using law of cosines for planar manuevers
    Delta V Law of Cosines

    (delta_v, alpha, beta) = delta_v_solve(v_1,v_2,fpa_1,fpa_2)

    Inputs: 
        - v_1 - velocity magnitude of initial orbit in km/sec
        - v_2 - velocity magnitude of final orbit in km/sec
        - fpa_1 - flight path angle of initial orbit in rad
        - fpa_2 - flight path angle of final orbit in rad

    Outputs: 
        - delta_v - delta_v required using the law of cosines
        - alpha - exterior angle of delta_v wrt to v_1 using law of sines
        - beta - interior angle between v1 and delta_v
    Dependencies: 
        - none

    Author: 
        - Shankar Kulumani 10 Oct 2012
        - Shankar Kulumani 30 Oct 2012
            - modified beta angle calculation
        - Shankar Kulumani 18 Nov 2017
            - Move to Python

    References
        - AAE532 notes LSN 19
        - AAE532_PS7.pdf
        - MAE3145 HW5 2017
    """
    delta_fpa = np.absolute(fpa2-fpa1)
    delta_v = np.sqrt(v2**2+v1**2-2*v2*v1*np.cos(delta_fpa))

    # need logic here to solve sign ambiguity
    sin_beta = v2*np.sin(delta_fpa)/delta_v
    cos_beta = (v2**2 - v1**2 - delta_v**2) / (-2 * delta_v * v1)
    beta = np.arctan2(sin_beta, cos_beta)
    alpha = np.pi-beta
    return delta_v, alpha, beta

# TODO: Add documentation and unit test
def planar_conic_orbit_intersection(p1, p2, ecc1, ecc2, dargp, nu1_old=np.deg2rad(90)):
    """HW6 Problem 3"""
    # Find nu1 for the original orbit
    
    def f(nu1):
        f = p2/(1+ecc2*np.cos(nu1-dargp)) - (p1/(1+ecc1*np.cos(nu1)))
        return f

    def fdot(nu1):
        fdot = ((p2*ecc2*np.sin(nu1-dargp))/(1+ecc2*np.cos(nu1-dargp))**2)-((p1*ecc1*np.sin(nu1))/(1+ecc1*np.cos(nu1))**2)
        return fdot
    nu1_new = fsolve(f, nu1_old)    
    # count = 0
    # delta = 1
    # while delta > 1e-9 and count < 50:
    #     count = count+1
        
    #     nu1_new = nu1_old - f(nu1_old)/fdot(nu1_old)
        
    #     delta = np.absolute(nu1_new - nu1_old)

    return nu1_new[0]

# TODO: Add unit tests
def delta_v_vnc(dv_mag, alpha, beta, fpa):
    """VNC LVLH Delta V vector

    ( dv_vnc, dv_lvlh ) = delta_v_vnc(dv_mag,alpha, beta, fpa)

    Purpose: 
        - Converts the alpha/beta angle of the Delta_V to the VNC frame
            - V - tangent to current velocity vector
            - N - normal to orbit plane (same direction as ang_mom vector)
            - C - orthongonal to velocity vector in orbit plane
        - Converts Delta_v into the LVLH frame
            - r_hat - in direction of orbit position vector
            - theta_hat - normal to r_hat in orbit plane
            - h_hat - orthogonal to orbit plane

    Inputs: 
        - dv_mag - magnitude of delta_v in km/sec
        - alpha - angle between projection of delta_v vector into the orbit
        plane and the orginal velocity vector (V) in rad
        - beta - angle of delta_v vector out of orbit plane (V-C plane) in
        rad
        - fpa - current flight path angle in rad - angle of the velocity
        vector wrt to theta_hat

    Outputs: 
        - dv_vnc - delta_v vector in VNC frame [V_hat;C_hat;N_hat];
        - dv_lvlh - delta_v vector in LVLH frame [r_hat;theta_hat;h_hat]

    Dependencies: 
        - none

    Author: 
        - Shankar Kulumani 15 Oct 2012
        - Shankar Kulumani 18 Nov 2017
            - In python

    References
        - AAE532 notes LSN 20
        - Vallado - referred to as NTW frame or Frenet System
    """

    dv_vnc = dv_mag*np.array([np.cos(beta)*np.cos(alpha), np.cos(beta)*np.sin(alpha), np.sin(beta)])

    phi = fpa+alpha

    dv_lvlh = dv_mag* np.array([np.cos(beta)*np.sin(phi),np.cos(beta)*np.cos(phi),np.sin(beta)])

    return dv_vnc, dv_lvlh

def hohmann(r_i,r_f,ecc_i,ecc_f,nu_i,nu_f,mu):
    """Hohmann Transfer between conic sections

    [dv_a,dv_b,tof ] = hohmann (r_i,r_f,ecc_i,ecc_f,nu_i,nu_f,mu)

    Inputs:
        - r_i - magnitude of initial position in km
        - r_f - magnitude of final position in km
        - ecc_i - eccentricity of initial orbit
        - ecc_f - eccentricity of final orbit
        - nu_i - true anomaly of initial orbit in rad (0 or pi)
        - nu_f - true anomaly of final orbit in rad (0 or pi)
        - mu - gravitational parameter of central body in km^3/sec^2

    Outputs:
        - dv_a - first delta_v in km/sec
        - dv_b - second delta_v in km/sec
        - tof - time of flight of transfer in seconds
        - phase_angle - angle in rad between departure point and arriving
        
    Dependencies:
        - tof_delta_t - calculates change in true anomaly given a TOF

    Author:
        - Shankar Kulumani 15 Oct 2012
        - Shankar Kulumani 17 Oct 2012
            - added phase angle
        - Shankar Kulumani 23 Nov 2017
            - move to python

    References
        - AAE532 Notes
        - Vallado
    """
    # calculate semi-major axis of initial,transfer, and final orbits
    a_i = (r_i * (1 + ecc_i * np.cos(nu_i))) / (1.0 - ecc_i**2 )
    a_f = (r_f * (1.0 + ecc_f * np.cos(nu_f))) / (1 - ecc_f**2 )

    a_t, p_t, ecc_t = kepler.perapo2aecc(r_i, r_f)
    p_i = kepler.semilatus_rectum(a_i, ecc_i)
    p_f = kepler.semilatus_rectum(a_f, ecc_f)
    # if ( ecc_i < 1.0 ) and ( ecc_f < 1.0 ):
    # find first delta_v
    v_i = vel_mag(r_i, a_i, mu)
    vt_a = vel_mag(r_i, a_t, mu)
    dv_a=  vt_a - v_i 
    
    # find second delta_v
    v_f = vel_mag(r_f, a_f, mu)
    vt_b = vel_mag(r_f, a_t, mu)
    dv_b=  v_f - vt_b 
    
    # ----------------  find transfer time of flight  ---------- }
    tof = np.pi * np.sqrt( a_t**3 / mu ) # always 1/2 period
    
    # find phase angle
    ( _,_ , phase_angle ) = kepler.tof_delta_t(p_f,ecc_f,mu,nu_f,-tof)
    
    return ( dv_a,dv_b,tof , phase_angle )    

def synodic_period(a1, a2, mu):
    r"""Compute synodic period between closed orbits

    S = synodic_period(a1, a2, mu)

    Parameters
    ----------
    a1 : float
        semimajor axis of smaller orbit
    a2 : float
        semimajor axis of larger orbit
    mu : float
        gravitational parameter of central body km^3/sec^2

    Returns
    -------
    S : float
        Synodic period in sec

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """
    n1 = np.sqrt(mu/a1**3)
    n2 = np.sqrt(mu/ a2**3)
    S = 2 * np.pi / np.absolute(n2 - n1)
    return S

def vel_mag(r, a, mu):
    r"""Find the velocity magnitude using the energy equation

    v = vel_mag(r, a, mu)

    Parameters
    ----------
    r : float
        radius of orbit in km
    a : float
        semimajor axis of orbit in km (negative for hyperbolic)
    mu : float
        gravitational parameter of central body km^3/sec^2

    Returns
    -------
    v : float
        velocity of orbit km/sec

    Notes
    -----
    You may include some math:

    .. math:: -\frac{\mu}{2 a} = \frac{v^2}{2} - \frac{\mu}{r}

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """ 
    v = np.sqrt((2 * mu / r) - mu / a)
    return v

# TODO: Add a phasing orbit function

# TODO: Add documentation and unit test
def simple_plane_change(v, del_inc):
    """Simple plane change delta v
    """

    delta_v = 2 * v * np.sin(del_inc/ 2)
    return delta_v


# TODO: Add combined plane change function 
# TODO: Update documentation and unit tests
def plane_change(v1, v2, delta_inc):
    """Solve for DeltaV using law of cosines for plane change

    (delta_v) = plane_change(v_1,v_2,delta_inc)

    Inputs: 
    """
    delta_v = np.sqrt(v2**2+v1**2-2*v2*v1*np.cos(delta_inc))
    return delta_v



#!/usr/bin/env python3

"""Satellite pass prediction module

This module implements PREDICT for command line use

Author
------
Shankar Kulumani		GWU		skulumani@gwu.edu
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import defaultdict
import numpy as np
from astro import geodetic, time, planets, tle, constants, transform, satellite
import os
import tempfile
import sys
import datetime
import argparse
import multiprocessing
import functools

deg2rad = constants.deg2rad
rad2deg = constants.rad2deg

mu = constants.earth.mu
re = constants.earth.radius
eesqrd = constants.earth.eesqrd
ee = constants.earth.ee

# TODO Add documentation and units for inputs
# TODO Setup logging
def predict(site_location, date_start, date_end, ifile='./tle.txt',
            ofile='./output.txt'):

    ifile = os.path.abspath(ifile)
    ofile = os.path.abspath(ofile)

    site_lat = np.deg2rad(site_location[0])
    site_lon = np.deg2rad(site_location[1])
    site_alt = float(site_location[2])
    
    site_ecef = geodetic.lla2ecef(site_lat, site_lon, site_alt)
    
    jd_start, _ = time.date2jd(date_start[0], date_start[1], date_start[2], date_start[3],
                               date_start[4], date_start[5])
    jd_end, _ = time.date2jd(date_end[0], date_end[1], date_end[2], date_end[3],
                               date_end[4], date_end[5])

    jd_step = 2 / (24 * 60) # step size in minutes
    jd_span = np.arange(jd_start, jd_end, jd_step)
    
    # echo check some data
    with open(ofile, 'w') as f:
        f.write('PREDICT RESULTS : Shankar Kulumani\n\n')
        f.write('Check Input Data : \n\n')
        f.write('Site Latitude    = {:9.6f} rad = {:9.6f} deg\n'.format(site_lat, np.rad2deg(site_lat)))
        f.write('Site Longitude   = {:9.6f} rad = {:9.6f} deg\n'.format(site_lon, np.rad2deg(site_lon)))
        f.write('Site Altitude    = {:9.6f} km  = {:9.6f} m\n'.format(site_alt, site_alt * 1000))
        
        f.write('\nObservation Window :\n\n')
        f.write('Start Date : {} UTC\n'.format(datetime.datetime(int(date_start[0]), int(date_start[1]), int(date_start[2]), int(date_start[3]), int(date_start[4]), int(date_start[5])).isoformat()))
        f.write('End Date   : {} UTC\n\n'.format(datetime.datetime(int(date_end[0]),int(date_end[1]),int(date_end[2]),int(date_end[3]),int(date_end[4]),int(date_end[5])).isoformat()))
        
        f.write('Start Julian Date   = {}\n'.format(jd_start))
        f.write('End Julian Date     = {}\n\n'.format(jd_end))

    # loop over jd span
    site = defaultdict(list)
    site['lat'] = site_lat
    site['lon'] = site_lon
    site['alt'] = site_alt
    site['ecef'] = site_ecef

    for jd in jd_span:
        gst, lst = time.gstlst(jd, site_lon)
        # site_eci = geodetic.site2eci(site_lat, site_alt, lst)
        # site_eci = attitude.rot3(gst).dot(site_ecef)
        site_eci = transform.dcm_ecef2eci(jd).dot(site_ecef)

        # test if site is in the dark
        sun_eci, _, _ = planets.sun_earth_eci(jd)

        # append site information to list
        site['eci'].append(site_eci)
        site['sun_eci'].append(sun_eci)
        site['gst'].append(gst)
        site['lst'].append(lst)

    # turn into numpy arrays
    for key, item in site.items():
        site[key] = np.squeeze(np.array(item))

    # update the downloaded TLEs if necessary

    # now we have to read the TLE
    sats = tle.get_tle(ifile)
    
    parallel_predict = functools.partial(satellite.parallel_predict, jd_span=jd_span, site=site)

    with multiprocessing.Pool(multiprocessing.cpu_count() - 1) as p:
        sats_passes = p.map(parallel_predict, sats)
    
    for sat, pass_vis in zip(sats, sats_passes):
        satellite.output(sat, pass_vis, ofile)

    # for sat in sats:
    #     sat.tle_update(jd_span)
    #     sat.visible(site)
    #     sat.output(ofile)

    return sats, sats_passes, site 

# TODO Add documentation
def parse_args(args):

    # default start and end times if none are given
    start_utc = list(datetime.datetime.today().timetuple()[0:6])
    end_utc = list((datetime.datetime.today() + datetime.timedelta(days=7)).timetuple()[0:6])
    
    # default output filename
    output_name = os.path.join(tempfile.gettempdir(),'predict_output.txt')

    # parse inputs
    # TODO Figure out command line arguments which should be positional vs optional
    # TODO Use logging in this funciton

    parser = argparse.ArgumentParser(description='PREDICT satellite passes')

    parser.add_argument('latitude', help='Latitude (deg) of Observation site', type=float)
    parser.add_argument('longitude', help='Longitude (deg) of Observation site', type=float)
    parser.add_argument('altitude', help='Altitude (km) of Observation site', type=float)

    parser.add_argument('--start', help='UTC Space separated list of year month day of start of prediction window.', 
                        default=start_utc, action='store', nargs=3, type=int)
    parser.add_argument('--end', help='UTC Space separated list of year month day of end of prediction window.', 
                        default=end_utc, action='store', nargs=3, type=int)
    # option to use saved tle
    parser.add_argument('-i', '--input', help='Path to input tle file', action='store', type=str)

    parser.add_argument('-o', '--output', help='Path to output file', action='store', type=str,
                        default=output_name)

    args = parser.parse_args(args)

    # option to download spacetrack tle and then run predict
    if not args.input:
        print('Using Celestrak visible sats')
        ifile = os.path.join(tempfile.gettempdir(),'predict_input.txt')
        tle.get_tle_visible(ifile)
    else:
        ifile = args.input
    
    return (args.latitude, args.longitude, args.altitude), args.start, args.end, ifile, args.output

if __name__ == "__main__":
    site_location, start, end, input_file, output_file = parse_args(sys.argv[1:])
    predict(site_location, start, end, input_file, output_file)





"""Test the TLE parsing and transformations
"""
import numpy as np
from astro import tle
import pdb


class TestTLEISS():
    l0 = "0 ISS (ZARYA)"
    l1 = "1 25544U 98067A   17168.12419852 +.00002055 +00000-0 +38402-4 0  9998"
    l2 = "2 25544 051.6427 038.6434 0004440 287.0211 207.7840 15.54043832061749"
    
    elements = tle.parsetle(l0, l1, l2)

    def test_validtle(self):
        np.testing.assert_equal(tle.validtle(self.l0, self.l1, self.l2), True)

    def test_line1_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l1), 8)

    def test_line2_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l2), 9)

    def test_satname(self):
        np.testing.assert_string_equal(self.elements.satname, "0 ISS (ZARYA)")

    def test_satnum(self):
        np.testing.assert_equal(self.elements.satnum, 25544)

    def test_classification(self):
        np.testing.assert_string_equal(self.elements.classification, 'U')

    def test_id_year(self):
        np.testing.assert_allclose(self.elements.id_year, 98)

    def test_id_launch(self):
        np.testing.assert_allclose(self.elements.id_launch, 67)

    def test_id_piece(self):
        np.testing.assert_string_equal(self.elements.id_piece, 'A  ')

    def test_epoch_year(self):
        np.testing.assert_allclose(self.elements.epoch_year, 17)

    def test_epoch_day(self):
        np.testing.assert_allclose(self.elements.epoch_day, 168.12419852)

    def test_ndot_over_2(self):
        np.testing.assert_allclose(self.elements.ndot_over_2, 0.00002055)
        
    def test_nddot_over_6(self):
        np.testing.assert_allclose(self.elements.nddot_over_6, 0)

    def test_bstar(self):
        np.testing.assert_allclose(self.elements.bstar, 38402e-4/1e5)

    def test_ephtype(self):
        np.testing.assert_allclose(self.elements.ephtype, 0)

    def test_elnum(self):
        np.testing.assert_allclose(self.elements.elnum, 999)

    def test_checksum1(self):
        np.testing.assert_allclose(self.elements.checksum1, 8)

    def test_inc(self):
        np.testing.assert_allclose(self.elements.inc, 51.6427)

    def test_raan(self):
        np.testing.assert_allclose(self.elements.raan, 38.6434)

    def test_ecc(self):
        np.testing.assert_allclose(self.elements.ecc, 0.0004440)

    def test_argp(self):
        np.testing.assert_allclose(self.elements.argp, 287.0211)

    def test_ma(self):
        np.testing.assert_allclose(self.elements.ma, 207.7840)

    def test_mean_motion(self):
        np.testing.assert_allclose(self.elements.mean_motion, 15.54043832)

    def test_epoch_rev(self):
        np.testing.assert_allclose(self.elements.epoch_rev, 6174)

    def test_checksum2(self):
        np.testing.assert_allclose(self.elements.checksum2, 9)


class TestTLESL3():
    l0 = "0 SL-3 R/B"
    l1 = "1 00056U 60011  B 60267.06075999  .41473555 +00000-0 +00000-0 0  9999"
    l2 = "2 00056 064.9400 008.3000 0007000 015.2810 344.7297 16.37352814005552"
    
    elements = tle.parsetle(l0, l1, l2)

    def test_validtle(self):
        np.testing.assert_equal(tle.validtle(self.l0, self.l1, self.l2), True)

    def test_line1_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l1), 9)

    def test_line2_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l2), 2)

    def test_satname(self):
        np.testing.assert_string_equal(self.elements.satname, "0 SL-3 R/B")

    def test_satnum(self):
        np.testing.assert_equal(self.elements.satnum, 56)

    def test_classification(self):
        np.testing.assert_string_equal(self.elements.classification, 'U')

    def test_id_year(self):
        np.testing.assert_allclose(self.elements.id_year, 60)

    def test_id_launch(self):
        np.testing.assert_allclose(self.elements.id_launch, 11 )

    def test_id_piece(self):
        np.testing.assert_string_equal(self.elements.id_piece, '  B')

    def test_epoch_year(self):
        np.testing.assert_allclose(self.elements.epoch_year, 60)

    def test_epoch_day(self):
        np.testing.assert_allclose(self.elements.epoch_day, 267.06075999)

    def test_ndot_over_2(self):
        np.testing.assert_allclose(self.elements.ndot_over_2, 0.41473555)
        
    def test_nddot_over_6(self):
        np.testing.assert_allclose(self.elements.nddot_over_6, 0)

    def test_bstar(self):
        np.testing.assert_allclose(self.elements.bstar, 0)

    def test_ephtype(self):
        np.testing.assert_allclose(self.elements.ephtype, 0)

    def test_elnum(self):
        np.testing.assert_allclose(self.elements.elnum, 999)

    def test_checksum1(self):
        np.testing.assert_allclose(self.elements.checksum1, 9)

    def test_inc(self):
        np.testing.assert_allclose(self.elements.inc, 64.9400)

    def test_raan(self):
        np.testing.assert_allclose(self.elements.raan, 8.300)

    def test_ecc(self):
        np.testing.assert_allclose(self.elements.ecc, 0.0007000)

    def test_argp(self):
        np.testing.assert_allclose(self.elements.argp, 15.2810)

    def test_ma(self):
        np.testing.assert_allclose(self.elements.ma, 344.7297)

    def test_mean_motion(self):
        np.testing.assert_allclose(self.elements.mean_motion, 16.373528140)

    def test_epoch_rev(self):
        np.testing.assert_allclose(self.elements.epoch_rev, 555)

    def test_checksum2(self):
        np.testing.assert_allclose(self.elements.checksum2, 2)

class TestTLEKOREASAT():
    l0 = "0 KOREASAT 5A"
    l1 = "1 42984U 17066A   17327.72549185 -.00000354  00000-0  00000+0 0  9990"
    l2 = "2 42984   0.0495 103.7501 0000074 356.8166 338.0747  1.00270863   278"
    
    elements = tle.parsetle(l0, l1, l2)

    def test_validtle(self):
        np.testing.assert_equal(tle.validtle(self.l0, self.l1, self.l2), True)

    def test_line1_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l1), 0)

    def test_line2_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l2), 8)

    def test_satname(self):
        np.testing.assert_string_equal(self.elements.satname, "0 KOREASAT 5A")

    def test_satnum(self):
        np.testing.assert_equal(self.elements.satnum, 42984)

    def test_classification(self):
        np.testing.assert_string_equal(self.elements.classification, 'U')

    def test_id_year(self):
        np.testing.assert_allclose(self.elements.id_year, 17)

    def test_id_launch(self):
        np.testing.assert_allclose(self.elements.id_launch, 66)

    def test_id_piece(self):
        np.testing.assert_string_equal(self.elements.id_piece, 'A  ')

    def test_epoch_year(self):
        np.testing.assert_allclose(self.elements.epoch_year, 17)

    def test_epoch_day(self):
        np.testing.assert_allclose(self.elements.epoch_day, 327.72549185)

    def test_ndot_over_2(self):
        np.testing.assert_allclose(self.elements.ndot_over_2, -0.00000354)
        
    def test_nddot_over_6(self):
        np.testing.assert_allclose(self.elements.nddot_over_6, 0)

    def test_bstar(self):
        np.testing.assert_allclose(self.elements.bstar, 0)

    def test_ephtype(self):
        np.testing.assert_allclose(self.elements.ephtype, 0)

    def test_elnum(self):
        np.testing.assert_allclose(self.elements.elnum, 999)

    def test_checksum1(self):
        np.testing.assert_allclose(self.elements.checksum1, 0)

    def test_inc(self):
        np.testing.assert_allclose(self.elements.inc, 0.0495)

    def test_raan(self):
        np.testing.assert_allclose(self.elements.raan, 103.7501)

    def test_ecc(self):
        np.testing.assert_allclose(self.elements.ecc, 0.0000074)

    def test_argp(self):
        np.testing.assert_allclose(self.elements.argp, 356.8166)

    def test_ma(self):
        np.testing.assert_allclose(self.elements.ma, 338.0747)

    def test_mean_motion(self):
        np.testing.assert_allclose(self.elements.mean_motion, 1.00270863)

    def test_epoch_rev(self):
        np.testing.assert_allclose(self.elements.epoch_rev, 27)

    def test_checksum2(self):
        np.testing.assert_allclose(self.elements.checksum2, 8)


class TestTLESL3():
    l0 = "0 SL-3 R/B"
    l1 = "1 00056U 60011  B 60267.06075999  .41473555 +00000-0 +00000-0 0  9999"
    l2 = "2 00056 064.9400 008.3000 0007000 015.2810 344.7297 16.37352814005552"
    
    elements = tle.parsetle(l0, l1, l2)

    def test_validtle(self):
        np.testing.assert_equal(tle.validtle(self.l0, self.l1, self.l2), True)

    def test_line1_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l1), 9)

    def test_line2_checksum(self):
        np.testing.assert_equal(tle.checksum(self.l2), 2)

    def test_satname(self):
        np.testing.assert_string_equal(self.elements.satname, "0 SL-3 R/B")

    def test_satnum(self):
        np.testing.assert_equal(self.elements.satnum, 56)

    def test_classification(self):
        np.testing.assert_string_equal(self.elements.classification, 'U')

    def test_id_year(self):
        np.testing.assert_allclose(self.elements.id_year, 60)

    def test_id_launch(self):
        np.testing.assert_allclose(self.elements.id_launch, 11 )

    def test_id_piece(self):
        np.testing.assert_string_equal(self.elements.id_piece, '  B')

    def test_epoch_year(self):
        np.testing.assert_allclose(self.elements.epoch_year, 60)

    def test_epoch_day(self):
        np.testing.assert_allclose(self.elements.epoch_day, 267.06075999)

    def test_ndot_over_2(self):
        np.testing.assert_allclose(self.elements.ndot_over_2, 0.41473555)
        
    def test_nddot_over_6(self):
        np.testing.assert_allclose(self.elements.nddot_over_6, 0)

    def test_bstar(self):
        np.testing.assert_allclose(self.elements.bstar, 0)

    def test_ephtype(self):
        np.testing.assert_allclose(self.elements.ephtype, 0)

    def test_elnum(self):
        np.testing.assert_allclose(self.elements.elnum, 999)

    def test_checksum1(self):
        np.testing.assert_allclose(self.elements.checksum1, 9)

    def test_inc(self):
        np.testing.assert_allclose(self.elements.inc, 64.9400)

    def test_raan(self):
        np.testing.assert_allclose(self.elements.raan, 8.300)

    def test_ecc(self):
        np.testing.assert_allclose(self.elements.ecc, 0.0007000)

    def test_argp(self):
        np.testing.assert_allclose(self.elements.argp, 15.2810)

    def test_ma(self):
        np.testing.assert_allclose(self.elements.ma, 344.7297)

    def test_mean_motion(self):
        np.testing.assert_allclose(self.elements.mean_motion, 16.373528140)

    def test_epoch_rev(self):
        np.testing.assert_allclose(self.elements.epoch_rev, 555)

    def test_checksum2(self):
        np.testing.assert_allclose(self.elements.checksum2, 2)



"""Coordinate transformations

"""

import numpy as np
from astro import kepler, constants, geodetic, time
from kinematics import attitude

# TODO: Add EOP and a full transformation here - compare to SPICE
def dcm_eci2ecef(jd):
    """Rotation matrix to convert from ECI to ECEF

    Will gradually improve this function to include all of the Earth
    motion terms. For now, just use sidereal time to get the rotation matrix
    """

    gst, _ = time.gstlst(jd, 0)

    dcm = attitude.rot3(gst, 'r')

    return dcm


def dcm_ecef2eci(jd):
    """Rotation matrix to transform from ECEF to ECI
    """
    dcm = dcm_eci2ecef(jd).T
    return dcm

# TODO: Documentation and unit testing
def dcm_pqw2eci_vector(r, v, mu=constants.earth.mu):
    """Define rotation matrix transforming PQW to ECI given eci vectors

    """
    h_vec, _, ecc_vec = kepler.hne_vec(r, v, mu)
    p_hat = ecc_vec / np.linalg.norm(ecc_vec)
    w_hat = h_vec / np.linalg.norm(h_vec) 
    q_hat = np.cross(w_hat, p_hat)

    dcm = np.stack((p_hat, q_hat, w_hat), axis=1)

    return dcm

#TODO: Documentation and testing
def dcm_pqw2eci_coe(raan, inc, arg_p):
    """Define rotation matrix transforming PQW to ECI given orbital elements

    """
    dcm = attitude.rot3(raan).dot(
            attitude.rot1(inc)).dot(attitude.rot3(arg_p))

    dcm_pqw2eci = attitude.rot3(-raan, 'r').dot(attitude.rot1(-inc, 'r')
                                                ).dot(attitude.rot3(-arg_p, 'r'))
    return dcm

def dcm_eci2pqw_coe(raan, inc, arg_p):
    """Define rotation matrix transforming PQW to ECI given orbital elements

    """
    dcm = dcm_pqw2eci_coe(raan, inc, arg_p).T
    return dcm

def dcm_pqw2lvlh(nu):
    dcm = attitude.rot3(-nu)
    return dcm

def dcm_lvlh2pqw(nu):
    dcm = dcm_pqw2lvlh(nu).T
    return dcm

#TODO: Documentation and testing
def dcm_sez2ecef(latgd, lon, alt=0):
    dcm = dcm_ecef2sez(latgd, lon)
    return dcm.T

#TODO: Documentation and testing
def dcm_ecef2sez(latgd, lon, alt=0):
    # dcm = attitude.rot2(np.pi/2 - latgd, 'r').dot(attitude.rot3(lon, 'r'))
    dcm = attitude.rot2(latgd - np.pi/2).dot(attitude.rot3(-lon))
    return dcm

#TODO: Documentation and testing
def dcm_ned2ecef(latgd, lon, alt=0):
    dcm = dcm_ecef2ned(latgd, lon, alt)
    return dcm.T

#TODO: Documentation and testing
def dcm_ecef2ned(latgd, lon, alt=0):
    dcm = np.array([[-np.sin(latgd)*np.cos(lon), -np.sin(lon), -np.cos(latgd)*np.cos(lon)],
                    [-np.sin(latgd)*np.sin(lon), np.cos(lon), -np.cos(latgd)*np.sin(lon)],
                    [np.cos(latgd), 0, -np.sin(latgd)]]).T
    return dcm

#TODO: Documentation and testing
def dcm_ecef2enu(latgd, lon, alt=0):
    dcm = np.array([[-np.sin(lon), np.cos(lon), 0],
                    [-np.sin(latgd)*np.cos(lon), -np.sin(latgd)*np.sin(lon), np.cos(latgd)],
                    [np.cos(latgd)*np.cos(lon), np.cos(latgd)*np.sin(lon), np.sin(latgd)]])

    return dcm

#TODO: Documentation and testing
def dcm_enu2ecef(latgd, lon, alt=0):
    dcm = dcm_ecef2enu(latgd, lon, alt)
    return dcm.T

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def enu2ecefv(enu, latgd, lon, alt=0):

    # find rotation matrix for this location enu2ecef
    dcm_enu2ecef = dcm_enu2ecef(latgd, lon)

    # vector to site (latgd, lon)
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    ecef = dcm_enu2ecef.dot(enu) + site_ecef

    return ecef

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def ecef2enuv(ecef, latgd, lon, alt=0):
    dcm_ecef2enu = dcm_ecef2enu(latgd, lon)

    # site vector
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    enu = dcm_ecef2enu.dot(ecef - site_ecef)
    return enu

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def ned2ecefv(ned, latgd, lon, alt=0):

    # find rotation matrix for this location enu2ecef
    dcm_ned2ecef = dcm_ned2ecef(latgd, lon)

    # vector to site (latgd, lon)
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    ecef = dcm_ned2ecef.dot(ned) + site_ecef

    return ecef

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def ecef2nedv(ecef, latgd, lon, alt=0):
    dcm_ecef2ned = dcm_ecef2ned(latgd, lon)

    # site vector
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    ned = dcm_ecef2ned.dot(ecef - site_ecef)
    return ned

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def sez2ecefv(sez, latgd, lon, alt=0):

    # find rotation matrix for this location enu2ecef
    dcm_sez2ecef = dcm_sez2ecef(latgd, lon)

    # vector to site (latgd, lon)
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    ecef = dcm_sez2ecef.dot(ned) + site_ecef

    return ecef

# TODO: Add documentation and testing
# TODO: Make this vectorized to allow for vector inputs
def ecef2sezv(ecef, latgd, lon, alt=0):
    dcm_ecef2sez = dcm_ecef2sez(latgd, lon)

    # site vector
    site_ecef = geodetic.lla2ecef(lat, lon, alt)

    # transform
    ned = dcm_ecef2sez.dot(ecef - site_ecef)
    return ned



"""Test out the predict module and function
"""

import numpy as np
from astro import predict
import os, pytest, filecmp


class TestPredict():
    
    @pytest.fixture(autouse=True)
    def setup(self, tmpdir):
        lat = 39.006
        lon = -104.883
        alt = 2.184
        
        site_location = (lat, lon, alt)
        start_date = (2006, 6, 19, 0, 0, 0)
        end_date = (2006, 6, 29, 0, 0, 0)
        
        self.input_file = './astro/tests/Predict.dat'
        self.output_file = os.path.join(str(tmpdir), 'output.txt')

        self.sats, self.sats_passes, self.site = predict.predict(site_location, start_date,
                                                end_date, self.input_file,
                                                self.output_file)
    # just check the first one
    
    def test_first_visible_pass(self):

        np.testing.assert_allclose(self.sats_passes[0][0].jd,
                                   (2453905.8333333,
                                    2453905.8347222,
                                    2453905.8361110))

    def test_output_file(self):
        np.testing.assert_equal(filecmp.cmp(self.output_file, './astro/tests/output.txt'), True)

class TestParserMinimumInput():
    
    latitude = 39.006
    longitude = -104.2
    altitude = 2.184

    site_location, start, end, input_file, output_file = predict.parse_args([str(latitude), str(longitude), str(altitude)])

    def test_latitude(self):
        np.testing.assert_allclose(self.site_location[0], self.latitude)

    def test_longitude(self):
        np.testing.assert_allclose(self.site_location[1], self.longitude)

    def test_altitude(self):
        np.testing.assert_allclose(self.site_location[2], self.altitude)







import numpy as np
import pdb
from astro import geodetic, time
from kinematics import attitude

deg2rad = np.pi/180

def test_equatorial_prime_meridian_lla2ecef():
    lat = 0 * np.pi/180
    lon = 0 * np.pi/180
    alt = 0
    expected_ecef = np.array([6378.137, 0, 0])
    actual_ecef = geodetic.lla2ecef(lat, lon, alt)
    np.testing.assert_array_almost_equal(actual_ecef, expected_ecef)

def test_equatorial_90_west_lla2ecef():
    lat = 0 * np.pi/180
    lon = 90 * np.pi/180
    alt = 0
    expected_ecef = np.array([0, 6378.137, 0])
    actual_ecef = geodetic.lla2ecef(lat, lon, alt)
    np.testing.assert_array_almost_equal(actual_ecef, expected_ecef)

def test_north_pole_lla2eccef():
    lat = 90 * np.pi/180
    lon = 0 * np.pi/180
    alt = 0
    expected_ecef = np.array([0, 0, 6356.7523142451792])
    actual_ecef = geodetic.lla2ecef(lat, lon, alt)
    np.testing.assert_array_almost_equal(actual_ecef, expected_ecef)

def test_lla2ecef_vallado():
    """Example 3-3 from Vallado
    """
    latgd = 34.352496 * np.pi/180
    lon = 46.4464 * np.pi/180
    hellp = 5085.22 # km
    expected_ecef = [6524.834, 6862.875, 6448.296] # km
    actual_ecef = geodetic.lla2ecef(latgd, lon, hellp)
    np.testing.assert_array_almost_equal(actual_ecef, expected_ecef, decimal=2)

def test_ecef2lla_vallado():
    """Example 3-3 from Vallado
    """
    ecef = [6524.834, 6862.875, 6448.296]
    # this is in latgd, lon, hellp in km
    expected_lla = [34.352496*np.pi/180, 46.4464 * np.pi/180, 5085.22]
    latgc, latgd, lon, hellp = geodetic.ecef2lla(ecef) 
    np.testing.assert_allclose([latgd, lon, hellp], expected_lla, rtol=1e-3)

def test_gc2gd_vallado():
    latgc = 34.173429 * np.pi/180
    expected_latgd = 34.352496 *np.pi/180
    actual_latgd = geodetic.gc2gd(latgc)
    np.testing.assert_allclose(actual_latgd, expected_latgd)

def test_gd2gc_vallado():
    latgd = 34.352496 * np.pi/180
    expected_latgc = 34.173429 * np.pi/180
    actual_latgc = geodetic.gd2gc(latgd)
    np.testing.assert_allclose(actual_latgc, expected_latgc)

def test_lla2ecef_usafa():
    lat = 39.006 * np.pi/180
    lon = -104.883 * np.pi/180
    alt = 2.184
    expected_ecef = [-1275.139, -4798.054, 3994.209]
    actual_ecef = geodetic.lla2ecef(lat, lon, alt)
    np.testing.assert_allclose(actual_ecef, expected_ecef, rtol=1e-3)

def test_lla2ecef_ascension_island():
    """Example 3-2 Vallado
    """
    latgd = np.deg2rad(-7.9066357)
    lon = np.deg2rad(345.5975)
    alt = 56 / 1e3
    ecef_expected = np.array([6119.40026932, -1571.47955545, -871.56118090])
    ecef = geodetic.lla2ecef(latgd, lon, alt)
    np.testing.assert_allclose(ecef, ecef_expected)

def test_lla2ecef_washington_dc():
    lat = 38.8895 * deg2rad
    lon = -77.0353 * deg2rad
    alt = 0.3
    expected_ecef = [1115.308, -4844.546, 3982.965]
    actual_ecef = geodetic.lla2ecef(lat, lon, alt)
    np.testing.assert_allclose(actual_ecef, expected_ecef, rtol=1e-3)

def test_ecef2lla2_washington_dc():
    ecef = [1115.308, -4844.546, 3982.965]
    expected_lla =  (38.8895 * deg2rad, 282.9647 * deg2rad, 0.3)
    latgc, latgd, longd, hellp  = geodetic.ecef2lla(ecef)
    np.testing.assert_allclose((latgd, longd, hellp), expected_lla, rtol=1e-3)

def test_rv2rhoazel_vallado():
    # test example 7-1 pg.409 Vallado
    latgd = np.deg2rad(39.007)
    lon = np.deg2rad(-104.883)
    alt = 2.187
    jd = 2449857.636829

    r = np.array([-5503.79562, 62.28191, 3824.24480])
    v = np.array([-2.199987, 1.139111, 1.484966])

    expected_output = (604.68, np.deg2rad(205.6), np.deg2rad(30.7), 2.08, np.deg2rad(0.15), np.deg2rad(0.17))  
    actual_output = geodetic.rv2rhoazel(r, v, latgd, lon, alt, jd)
    np.testing.assert_allclose(actual_output, expected_output, rtol=1e-4)

def test_rv2rhoazel_vallado():
    # Example 4-1 pg. 275 Vallado
    latgd = np.deg2rad(39.007)
    lon = np.deg2rad(-104.883)
    alt = 2.19456 # kilometer
    jd, _ = time.date2jd(1994, 5, 14, 13, 11, 20.59856)

    # ECI values for position and velocity
    r_eci = np.array([1752246215, -3759563433, -1577568105])
    v_eci = np.array([-18.323, 18.332, 7.777])
    
    expected_output = (4437722626.456, 3.67834, 0.4171786,
                       -25.360829, 6.748139e-5,-2.89766e-5)
    actual_output = geodetic.rv2rhoazel(r_eci, v_eci, latgd, lon, alt, jd)
    np.testing.assert_allclose(actual_output, expected_output, rtol=1e-3)

def test_rhoazel_equator_zenith():
    sat_eci = np.array([6378.137 + 100, 0, 0])
    site_eci = np.array([6378.137, 0, 0])
    site_lat = 0
    site_lst = 0
    
    true_rho = 100
    true_az = np.pi
    true_el = np.pi /2

    rho, az, el = geodetic.rhoazel(sat_eci, site_eci, site_lat, site_lst)
    np.testing.assert_allclose((rho, az, el), (true_rho, true_az, true_el))

def test_rhoazel_pole_zenith():
    sat_eci = np.array([0, 0, 6378.137 + 200])
    site_eci = np.array([0, 0, 6378.137])
    site_lat = np.pi / 2
    site_lst = 0
    
    true_rho = 200
    true_az = np.pi
    true_el = np.pi / 2

    rho, az, el = geodetic.rhoazel(sat_eci, site_eci, site_lat, site_lst)
    np.testing.assert_allclose((rho, az, el), (true_rho, true_az, true_el))


class TestRhoAzEl2SEZ():

    def test_zenith_stationary(self):
        rho = np.random.uniform(0, 1000)
        az = 0
        el = np.pi/2
        drho = 0
        daz = 0
        dele = 0
        rho_sez_expected = np.array([0, 0, rho])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)

    def test_horizon_north(self):
        rho = np.random.uniform(0, 1000)
        az = 0
        el = 0
        drho = 0
        daz = 0
        dele = 0
        rho_sez_expected = np.array([-rho, 0, 0])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)

    def test_horizon_east(self):
        rho = np.random.uniform(0, 1000)
        az = np.pi/2
        el = 0
        drho = 0
        daz = 0
        dele = 0
        rho_sez_expected = np.array([0, rho, 0])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)

    def test_zenith_approaching(self):
        rho = np.random.uniform(0, 1000)
        az = 0
        el = np.pi/2
        drho = -np.random.uniform(0, 100)
        daz = 0
        dele = 0
        rho_sez_expected = np.array([0, 0, rho])
        drho_sez_expected = np.array([0, 0, drho])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)
        np.testing.assert_array_almost_equal(drho_sez, drho_sez_expected)

    def test_horizon_north_receeding(self):
        rho = np.random.uniform(0, 1000)
        az = 0
        el = 0
        drho = np.random.uniform(0, 10)
        daz = 0
        dele = 0
        rho_sez_expected = np.array([-rho, 0, 0])
        drho_sez_expected = np.array([-drho, 0, 0])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)
        np.testing.assert_array_almost_equal(drho_sez, drho_sez_expected)

    def test_horizon_east_approaching(self):
        rho = np.random.uniform(0, 1000)
        az = np.pi/2
        el = 0
        drho = -np.random.uniform(0, 100)
        daz = 0
        dele = 0
        rho_sez_expected = np.array([0, rho, 0])
        drho_sez_expected = np.array([0, drho, 0])
        rho_sez, drho_sez = geodetic.rhoazel2sez(rho, az, el, drho, daz, dele)
        np.testing.assert_array_almost_equal(rho_sez, rho_sez_expected)
        np.testing.assert_array_almost_equal(drho_sez, drho_sez_expected)
